#!/usr/bin/python3
# -*- coding: utf-8 -*-
# Author: i2cy(i2cy@outlook.com)
# Project: I2FTPServer
# Filename: process_manage
# Created on: 2022/5/17

import os
import signal
import time
from i2cylib.network.I2TCP import Client
from i2cylib.utils.logger import Logger
from i2cylib.utils import args
from i2ftps.config import Config


LOGGER = Logger(level="INFO")
KILL_TIMEOUT = 20


def stop(config):
    # 连接到服务器
    assert isinstance(config, Config)
    LOGGER.INFO("[Main] stopping target server".format(config.port))
    LOGGER.INFO("[Main] connecting to target server at port {}".format(config.port))
    clt = Client("localhost", config.port, config.keychain, logger=LOGGER)

    try:
        clt.connect()
        flag = clt.get(timeout=5)
        if not "I2FTP" in flag.decode("utf-8"):
            raise Exception("target is not a I2FTP server")
    except Exception as err:
        LOGGER.ERROR("[Main] failed to connect to server, {}".format(err))
        return

    # 发送PCTL命令
    try:
        clt.send(b"PCTL")
    except Exception as err:
        LOGGER.ERROR("[Main] server respond incorrectly, {}".format(err))
        return

    # 发送终止信号
    try:
        feedback = clt.get(timeout=5)
        ret, pid = feedback.split(b",", 1)
        pid = int().from_bytes(pid, "little", signed=False)
        if os.name == 'nt':
            os.system("taskkill -f -t -pid {}".format(pid))
        else:
            os.kill(pid, signal.SIGINT)
            t0 = time.time()
            while True:
                if time.time() - t0 > KILL_TIMEOUT:
                    LOGGER.WARNING("[Main] process may not be killed, timeout")
                    break
    except Exception as err:
        LOGGER.ERROR("[Main] failed to decode feedback from server or failed to kill process, {}".format(
            err
        ))



#!/usr/bin/python3
# -*- coding: utf-8 -*-
# Author: i2cy(i2cy@outlook.com)
# Project: I2FTPServer
# Filename: client
# Created on: 2022/5/24

import os
import time
import json
import tqdm
import hashlib
import pathlib
from i2cylib.network import Client
from i2cylib.utils import Logger
from i2ftps.server import VERSION


class I2ftpClient:

    def __init__(self, hostname, key=b"&90%]>__AdfI2FTP$F%_+@$^:aBasicKey%_+@-$^:>",
                 port=26842, logger=None, max_buffer_size=1000, timeout=15):
        self.version = b"I2FTP " + VERSION.encode("utf-8")
        if logger is None:
            logger = Logger()
        self.logger = logger
        self.timeout = timeout
        self.i2clt = Client(hostname, port, key, logger=logger, max_buffer_size=max_buffer_size)

        self.__header = "[I2FTP]"
        self.flag_download_busy = False

    def __del__(self):
        self.i2clt.reset()

    def send_command(self, cmd, feedback=True):
        self.i2clt.send(cmd)

        if feedback:
            status, ret = self.get_feedback()

            return status, ret

    def get_feedback(self):
        feed = self.i2clt.get(timeout=self.timeout)

        status = False
        ret = "timeout when receiving feedback"

        if feed is not None:
            status, ret = feed.split(b",", 1)
            status = bool(status[0])

        if not status:
            ret = ret.decode("utf-8")

        return status, ret

    def connect(self, timeout=None):
        if timeout is None:
            timeout = self.timeout
        else:
            self.timeout = timeout
        err = ""
        self.i2clt.connect(timeout=timeout)
        version = self.i2clt.get(timeout=timeout)
        if version != self.version:
            err = "failed to match server version, server: {} client: {}".format(
                version, self.version
            )
            self.logger.ERROR("{} ".format(self.__header, err))
            self.i2clt.reset()
        return self.i2clt.connected, err

    def disconnect(self):
        return self.i2clt.reset()

    def list(self, path):
        cmd = "LIST,{}".format(path).encode("utf-8")

        status, ret = self.send_command(cmd)

        if status:
            ret = ret.decode("utf-8")
            ret = json.loads(ret)

        return status, ret

    def download(self, path):
        cmd = "GETF,{}".format(path).encode("utf-8")

        # 获得会话ID
        status, ret = self.send_command(cmd)

        if status:
            session_id = ret
            # 获得文件信息
            status, file_details = self.list(path)
            if not status:
                return status, file_details
            assert isinstance(file_details, dict)
            file_details = file_details[pathlib.Path(path).name]
            # 新建下载会话（新连接）
            ret = DownloadSession(session_id, file_details,
                                  self.i2clt.address[0], key=self.i2clt.key, port=self.i2clt.address[1],
                                  logger=self.logger, max_buffer_size=self.i2clt.max_buffer,
                                  timeout=self.timeout)
            status, res = ret.connect()
            if not status:
                return status, res

        return status, ret


class DownloadSession(I2ftpClient):

    def __init__(self, session_id, file_details, hostname,
                 key=b"&90%]>__AdfI2FTP$F%_+@$^:aBasicKey%_+@-$^:>",
                 port=26842, logger=None, max_buffer_size=1000, timeout=15):
        
        super(DownloadSession, self).__init__(hostname, key, port, logger, max_buffer_size, timeout)
        self.session_id = session_id
        self.details = file_details
        self.closed = False
        self.hash_object = hashlib.md5()
        self.length = file_details["size"]

    def __del__(self):
        if not self.closed:
            cmd = b"CLOZ," + self.session_id
            self.send_command(cmd, feedback=False)
        super(DownloadSession, self).__del__()

    def __len__(self):
        return self.length

    def __getitem__(self, item):
        if self.flag_download_busy:
            raise Exception("client download busy, only one downloading quest a time")
        if self.closed:
            raise Exception("session closed")

        if isinstance(item, slice):
            cmd = b"DOWN," + self.session_id + b","

            start = item.start
            if start < 0:
                start = self.length - start
            stop = item.stop
            if stop < 0:
                stop = self.length - stop

            cmd += int(start).to_bytes(8, "little", signed=False) + b","
            cmd += int(stop).to_bytes(8, "little", signed=False)
            length = stop - start
            dat = b""
            self.flag_download_busy = True
            self.send_command(cmd, feedback=False)

            while len(dat) < length:
                status, fed = self.get_feedback()
                if not status:
                    raise Exception(fed)
                if start + len(dat) != int().from_bytes(dat[:8], "little", signed=False):
                    raise Exception("expecting package loss, received index {} (expecting {})".format(
                        int().from_bytes(dat[:8], "little", signed=False), start + len(dat)
                    ))
                dat += fed[9:]
            ret = dat[::item.step]

        else:
            cmd = b"DOWN," + self.session_id + b","
            if item < 0:
                item = self.length - item
            cmd += int(item).to_bytes(8, "little", signed=False) + b","
            cmd += int(item + 1).to_bytes(8, "little", signed=False)
            status, ret = self.send_command(cmd)
            if not status:
                self.flag_download_busy = False
                raise Exception(ret)
            if item != int().from_bytes(ret[:8], "little", signed=False):
                self.flag_download_busy = False
                raise Exception("expecting package loss, received index {} (expecting {})".format(
                    int().from_bytes(ret[:8], "little", signed=False), item
                ))
            ret = ret[9:]

        self.flag_download_busy = False
        return ret

    def close(self):
        if self.closed:
            return self.closed, -1
        cmd = b"CLOZ," + self.session_id
        status, ret = self.send_command(cmd)
        if status:
            self.closed = True
            ret = ret.decode("utf-8")
            ret = self.hash_object.hexdigest() == ret

        super(DownloadSession, self).disconnect()

        return status, ret

    def to_file(self, filename, verbose=True, close_session_when_done=True):
        if self.flag_download_busy:
            raise Exception("client download busy, only one downloading quest a time")
        if self.closed:
            raise Exception("session closed")
        self.hash_object = hashlib.sha256()
        f = open(filename, "wb")

        received = 0

        cmd = b"DOWN," + self.session_id + b","
        cmd += b"\x00" * 8 + b","
        cmd += int(self.length).to_bytes(8, "little", signed=False)
        self.flag_download_busy = True
        self.send_command(cmd, feedback=False)

        status, hash_res = False, -1
        if verbose:
            pbar = tqdm.tqdm(desc=pathlib.Path(filename).name, total=self.length,
                             unit="B", unit_scale=True)

        while received < self.length:
            status, fed = self.get_feedback()
            if not status:
                raise Exception(fed)
            if received != int().from_bytes(fed[:8], "little", signed=False):
                if verbose:
                    pbar.close()
                raise Exception("expecting package loss, received index {} (expecting {})".format(
                    int().from_bytes(fed[:8], "little", signed=False), received
                ))
            dat = fed[9:]
            self.hash_object.update(dat)
            received += f.write(dat)
            if verbose:
                pbar.update(len(dat))

        if verbose:
            pbar.close()

        if close_session_when_done:
            status, hash_res = self.close()
        else:
            status = True

        return status, hash_res


if __name__ == '__main__':
    test_server = "i2cy.tech"
    test_port = 26842
    test_key = b"&90%]>__AdfI2FTP$F%_+@$^:aBasicKey%_+@-$^:>"
    test_file = "small.mp4"
    test_log = "test.log"

    clt = I2ftpClient(test_server, test_key, test_port, logger=Logger(test_log, echo=False))

    clt.connect()
    print("test server connected")

    print("<-> listing root files")
    state, data = clt.list(".")
    print("<*> LIST test result: {}".format(state))
    print("    files under root: \n{}".format(json.dumps(data, indent=2)))

    print("<-> downloading test file {}".format(test_file))
    state, session = clt.download(test_file)
    print("<*> GETF test result: {}".format(state))
    if isinstance(session, DownloadSession):
        state, md5_res = session.to_file(test_file, True)
    else:
        state, md5_res = False, False
    print("<*> DOWN test result: {}".format(state))
    print("<*> file hash matching result: {}".format(md5_res))

    clt.disconnect()
    print("disconnected from server")
    print("test ended")


#!/usr/bin/python3
# -*- coding: utf-8 -*-
# Author: i2cy(i2cy@outlook.com)
# Project: I2FTPServer
# Filename: server
# Created on: 2022/5/12


import threading
import queue
import time
import hashlib
import os
import pathlib
import json
import shutil
from i2cylib.network.I2TCP import Server, Handler
from i2cylib.utils.logger import Logger
from i2cylib.utils.path import path_fixer
from i2cylib.utils.bytes import random_keygen
from i2cylib.utils.args import get_args
from i2ftps.config import Config
import i2ftps.process_manage as pm

TIMEOUT = 20
MAX_CONNECTIONS = 100
MAX_UPDOWN_SESSIONS = 500
MAX_CMD_QUEUE = 1000
MAX_PACKAGE_BUFFER = 100
FILE_SESSION_TIMEOUT = 120
FULL_SPEED_TIMEOUT = 0.5

SIGNAL_PACKAGE_SIZE = 524128

VERSION = "1.0"

DEBUG = True


class FileSession:

    def __init__(self, path, readonly=True):
        """
        file I/O session object

        :param path: Path like object
        :param readonly: bool, set session to readonly
        """
        if readonly:
            self.io = open(path, "rb")
            self.__hash_io = open(path, "rb")
        else:
            self.io = open(path, "wb+")
            self.__hash_io = self.io
        self.readonly = readonly
        self.__md5 = hashlib.md5()
        self.size = os.path.getsize(path)
        self.flag_md5_available_oneshot = False
        self.__flag_md5_available = False
        self.__last_io_ts = time.time()
        self.closed = False

    def seek(self, offset):
        """
        change current file pointer to given offset

        :param offset: int
        :return: None
        """
        if self.io.tell() == offset:
            return

        self.__last_io_ts = time.time()

        self.io.seek(offset)

    def write(self, data):
        """
        write data to standard file I/O

        :param data: bytes
        :return: int, length
        """
        self.__last_io_ts = time.time()

        try:
            ret = self.io.write(data)
        except Exception as err:
            ret = 0

        return ret

    def read(self, length):
        """
        read data from standard file I/O

        :param length: int, max data length
        :return: bytes, data
        """
        self.__last_io_ts = time.time()

        ret = self.io.read(length)

        return ret

    def calc_hash(self, step=True, size=16384):
        """
        calculate file's md5 sum value by step or instantly

        :param step: bool, calculate one step
        :param size: int, read size for one step
        :return: bool, calculation status (False for already done the calculation)
        """
        if self.__flag_md5_available:
            return False

        if step:
            data = self.__hash_io.read(size)
            length = len(data)
            if length < size:
                self.flag_md5_available_oneshot = True
                self.__flag_md5_available = True
                self.__hash_io.close()
            self.__md5.update(data)
        else:
            while True:
                data = self.__hash_io.read(size)
                if not data:
                    break
                self.__md5.update(data)
            self.flag_md5_available_oneshot = True
            self.__flag_md5_available = True
            self.__hash_io.close()

        return True

    def age(self):
        """
        return the time (s) of this session from the last interact

        :return: float, time in seconds
        """
        return time.time() - self.__last_io_ts

    def md5(self):
        """
        return file's md5 sum, if calculation is still in progress,
        status will be False

        :return: (bool status, hex_str sum)
        """
        self.__last_io_ts = time.time()
        if self.__flag_md5_available:
            return True, self.__md5.hexdigest()
        else:
            return False, None

    def close(self):
        """
        close session

        :return: None
        """
        self.io.close()
        self.closed = True


class I2ftpServer:

    def __init__(self, config):
        """
        I2FTP object

        :param config: i2ftps.Config
        """

        assert isinstance(config, Config)

        self.__header = "[I2FTP]"
        self.__pid = os.getpid()
        self.config = config
        self.__server = None
        self.logger = Logger(config.log_file,
                             level=config.log_level)
        self.root = pathlib.Path(config.ftp_root)
        if not self.root.is_absolute():
            self.logger.CRITICAL("{} [Init] root path has to be absolute path")
            raise Exception("root path in configuration has to be absolute path")
        if not self.root.exists():
            path_fixer(config.ftp_root)

        self.__flag_kill = False
        self.threads_running = {"loop": False,
                                "file_session_manage_loop": False}

        self.connections = []  # handler池
        self.__cmd_queue = queue.Queue(maxsize=MAX_CMD_QUEUE)  # 命令池

        self.__file_session = {}  # 文件会话池
        self.__file_session_queue = []  # 文件会话缓冲

    def start(self):
        """
        启动I2FTP服务器

        :return: None
        """
        self.__server = Server(key=self.config.keychain,
                               port=self.config.port,
                               logger=self.logger,
                               max_con=MAX_CONNECTIONS,
                               timeout=TIMEOUT,
                               secured_connection=self.config.tls_enabled,
                               max_buffer_size=MAX_PACKAGE_BUFFER
                               )
        self.__server.start()
        threading.Thread(target=self.__loop).start()
        threading.Thread(target=self.__file_session_manage_loop).start()
        self.logger.INFO("{} I2FTP server started".format(self.__header))

    def stop(self):
        """
        停止I2FTP服务器

        :return:
        """
        assert isinstance(self.__server, Server)
        self.__flag_kill = True
        keep = True
        t0 = time.time()
        while keep:
            time.sleep(0.02)
            keep = False
            for ele in self.threads_running.keys():
                if self.threads_running[ele]:
                    keep = True
                    break
            if time.time() - t0 > TIMEOUT:
                self.logger.CRITICAL("{} [Main] threads not stopping, timeout, {}".format(
                    self.__header, self.threads_running))
                break

        self.__server.kill()
        self.logger.INFO("{} I2FTP server stopped".format(self.__header))

    def __check_path(self, path):
        # 检查路径是否符合安全规定
        raw = path.replace("\\", "/").split("/")

        if ".." in raw or not raw[0]:
            return False, b"\x00,requesting parent directory or using absolute path is not allowed"

        return True, b""

    def __process_requests(self, requests):
        assert isinstance(requests, bytes)
        header = "[CmdProcess]"
        requests = requests.split(b",", 1)
        cmd = requests[0]
        payload = requests[1]

        ret = b""
        ret_cmd = b""

        if cmd == b"LIST":  # 查询命令
            res = {}
            path = payload.decode("utf-8")
            # 检查路径是否符合安全规定
            status, ret = self.__check_path(path)
            if not status:
                return ret, ret_cmd
            path = self.root.joinpath(path)

            # 检查路径是否存在
            if not path.exists():
                return b"\x00,path requested does not exist", ret_cmd

            # 当路径是文件时
            if path.is_file():
                res.update({
                    path.name: {
                        "is_dir": False,
                        "size": os.path.getsize(path),
                        "time": os.path.getmtime(path)
                    }
                })

            # 当路径是目录时
            else:
                path = [ele for ele in path.glob("*")]
                for ele in path:
                    res.update({
                        ele.name: {
                            "is_dir": ele.is_dir(),
                            "size": os.path.getsize(ele),
                            "time": os.path.getmtime(ele)
                        }
                    })
            ret = b"\x01," + json.dumps(res).encode("utf-8")

        elif cmd == b"GETF":  # 创建下载会话指令
            # 检查会话池是否已满
            if len(self.__file_session) >= MAX_UPDOWN_SESSIONS:
                return b"\x00,file session full on server", ret_cmd

            # 检查文件路径是否符合要求
            path = payload.decode("utf-8")
            status, ret = self.__check_path(path)
            if not status:
                return ret, ret_cmd
            path = self.root.joinpath(path)

            # 检查文件是否存在且是文件
            if not path.exists() or path.is_dir():
                return b"\x00,path does not exist or target is a directory", ret_cmd

            # 建立会话
            session = FileSession(path, readonly=True)

            session_id = random_keygen(16)
            while session_id in self.__file_session:
                session_id = random_keygen(16)

            self.logger.DEBUG("{} {} download session created (session ID: {})".format(
                self.__header, header, session_id.hex()
            ))

            self.__file_session_queue.append((session_id, session))

            ret = b"\x01," + session_id

        elif cmd == b"DOWN":  # 通过会话下载文件
            # 分割指令
            session_id = payload[:16]
            fp = int().from_bytes(payload[17:25], "little", signed=False)
            fp_end = int().from_bytes(payload[26:34], "little", signed=False)

            # 检查会话是否有效
            if session_id not in self.__file_session:
                return b"\x00,invalid session id quested", ret_cmd

            session = self.__file_session[session_id]
            assert isinstance(session, FileSession)

            session.seek(fp)

            if fp_end <= fp:
                fp_end = session.size
            length = fp_end - fp
            if length > SIGNAL_PACKAGE_SIZE:  # 当请求的长度大于单个包能承载的长度时
                data = session.read(SIGNAL_PACKAGE_SIZE)
                ret_cmd = b"DOWN," + session_id + b","
                ret_cmd += int(fp + SIGNAL_PACKAGE_SIZE).to_bytes(8, "little", signed=False) + b","
                ret_cmd += payload[26:34]
            else:
                data = session.read(length)

            ret = b"\x01," + payload[17:25] + b"," + data

        elif cmd == b"PULF":  # 创建上传会话命令
            # 若服务器只读则拒绝
            if self.config.read_only:
                return b"\x00,read-only server", ret_cmd

            # 检查会话池是否已满
            if len(self.__file_session) >= MAX_UPDOWN_SESSIONS:
                return b"\x00,file session full on server", ret_cmd

            # 检查文件路径是否符合要求
            path = payload.decode("utf-8")
            status, ret = self.__check_path(path)
            if not status:
                return ret, ret_cmd
            path = self.root.joinpath(path)

            # 检查路径是否是文件夹
            if path.is_dir():
                return b"\x00,path does not exist or target is a directory", ret_cmd

            # 若路径不存在则创建路径
            path_fixer(path.as_posix())

            # 建立会话
            session = FileSession(path, readonly=False)

            session_id = random_keygen(16)
            while session_id in self.__file_session:
                session_id = random_keygen(16)

            self.__file_session_queue.append((session_id, session))
            self.logger.DEBUG("{} {} upload session created (session ID: {})".format(
                self.__header, header, session_id.hex()
            ))

            ret = b"\x01," + session_id

        elif cmd == b"UPLD":  # 通过会话上传数据
            # 若服务器只读则拒绝
            if self.config.read_only:
                return b"\x00,read-only server", ret_cmd

            # 分割指令
            session_id = payload[:16]
            fp = int().from_bytes(payload[17:25], "little", signed=False)
            data = payload[26:]

            # 检查会话是否有效
            if session_id not in self.__file_session:
                return b"\x00,invalid session id quested", ret_cmd

            session = self.__file_session[session_id]
            assert isinstance(session, FileSession)

            # 检查会话是否是上传会话
            if session.readonly:
                return b"\x00,read-only session", ret_cmd

            # 储存数据
            session.seek(fp)
            fp += session.write(data)
            ret = b"\x01," + fp.to_bytes(8, "little", signed=False)

        elif cmd == b"CLOZ":  # 关闭会话命令
            session_id = payload

            # 检查会话是否有效
            if session_id not in self.__file_session:
                return b"\x00,invalid session id quested", ret_cmd

            session = self.__file_session[session_id]
            assert isinstance(session, FileSession)

            # 若会话为下载会话
            if session.readonly:
                status, md5 = session.md5()

                # 检查md5值是否运算完毕
                if not status:
                    return b"\x00,session is still calculating file's md5 sum", ret_cmd

                # 关闭会话
                session.close()
                ret = b"\x01," + md5.encode("utf-8")

            # 若为上传会话
            else:
                # 关闭会话
                session.close()
                ret = b"\x01"

            self.logger.DEBUG("{} {} file session closed (session ID: {})".format(
                self.__header, header, session_id.hex()
            ))

        elif cmd == b"FIOP":  # 文件操作命令
            # 若服务器只读则拒绝
            if self.config.read_only:
                return b"\x00,read-only server", ret_cmd

            # 分割命令
            command = payload[0]
            args = payload[2:].decode("utf-8")
            paths = args.split(",")

            # 检查文件路径是否符合要求
            for path in paths:
                status, ret = self.__check_path(path)
                if not status:
                    return ret, ret_cmd

            # 重命名
            if command == 0:
                path0 = self.root.joinpath(paths[0])
                path1 = self.root.joinpath(paths[1])

                # 检查路径是否存在
                if not path0.exists():
                    return b"\x00,path requested does not exist", ret_cmd
                if path1.exists():
                    return b"\x00,target name already exists", ret_cmd

                try:
                    os.rename(path0, path1)
                except Exception as err:
                    return b"\x00," + str(err).encode("utf-8"), ret_cmd

            # 移动
            elif command == 1:
                path0 = self.root.joinpath(paths[0])
                path1 = self.root.joinpath(paths[1])

                # 检查路径是否存在
                if not path0.exists():
                    return b"\x00,path requested does not exist", ret_cmd
                if path1.exists():
                    return b"\x00,target path already exists", ret_cmd

                # 移动文件
                try:
                    shutil.move(path0, path1)
                except Exception as err:
                    return b"\x00," + str(err).encode("utf-8"), ret_cmd

            # 复制
            elif command == 2:
                path0 = self.root.joinpath(paths[0])
                path1 = self.root.joinpath(paths[1])

                # 检查路径是否存在
                if not path0.exists():
                    return b"\x00,path requested does not exist", ret_cmd
                if path1.exists():
                    return b"\x00,target path already exists", ret_cmd

                # 复制文件、文件夹
                try:
                    if path0.is_file():
                        shutil.copyfile(path0, path1)
                    else:
                        shutil.copytree(path0, path1)
                except Exception as err:
                    return b"\x00," + str(err).encode("utf-8"), ret_cmd

            # 删除
            elif command == 3:
                path0 = self.root.joinpath(paths[0])

                # 检查路径是否存在
                if not path0.exists():
                    return b"\x00,path requested does not exist", ret_cmd

                # 删除文件、文件夹
                try:
                    if path0.is_file():
                        os.remove(path0)
                    else:
                        shutil.rmtree(path0)
                except Exception as err:
                    return b"\x00," + str(err).encode("utf-8"), ret_cmd

            # 创建文件夹
            elif command == 4:
                path0 = self.root.joinpath(paths[0])

                # 检查路径是否存在
                if path0.exists():
                    return b"\x00,target path already exists", ret_cmd

                # 创建文件夹
                try:
                    os.makedirs(path0)
                except Exception as err:
                    return b"\x00," + str(err).encode("utf-8"), ret_cmd

            ret = b"\x01"

        return ret, ret_cmd

    def __file_session_manage_loop(self):
        if self.threads_running["file_session_manage_loop"]:
            return
        self.threads_running["file_session_manage_loop"] = True

        full_speed_ts = time.time()
        header = "[SessionManage]"
        idle = False

        while not self.__flag_kill:
            pops = []
            for session_id, session in self.__file_session_queue[:]:
                self.__file_session.update({session_id: session})
                self.__file_session_queue.pop(0)
            for ele in self.__file_session.keys():
                session = self.__file_session[ele]
                assert isinstance(session, FileSession)

                if self.__flag_kill:
                    break

                # 自动关闭不活动的会话
                if session.age() > FILE_SESSION_TIMEOUT:
                    session.close()
                    pops.append(ele)
                    self.logger.WARNING("{} {} inactive file session found: \"{}\", closing".format(
                        self.__header, header, ele.hex()
                    ))
                    continue

                # 关闭死亡的会话
                if session.closed:
                    pops.append(ele)

                # 计算会话文件的md5校验和
                ret = session.calc_hash(step=True)
                if ret:
                    full_speed_ts = time.time()
                else:
                    if session.flag_md5_available_oneshot:
                        self.logger.DEBUG("{} {} md5 value of session \"{}\" calculated".format(
                            self.__header, header, ele.hex()
                        ))
                        session.flag_md5_available_oneshot = False

            if time.time() - full_speed_ts > FULL_SPEED_TIMEOUT:
                time.sleep(0.01)
                if not idle:
                    idle = True
                    self.logger.DEBUG("{} {} thread idle".format(self.__header, header))
            else:
                idle = False

            for ele in pops:
                self.__file_session.pop(ele)

        self.threads_running["file_session_manage_loop"] = False

    def __loop(self):
        assert isinstance(self.__server, Server)
        if self.threads_running["loop"]:
            return
        self.threads_running["loop"] = True

        full_speed_ts = time.time()
        cnt = 0
        header = "[MainLoop]"
        idle = False

        while not self.__flag_kill:
            cnt += 1
            if cnt > 9999:
                cnt = 0

            if time.time() - full_speed_ts > FULL_SPEED_TIMEOUT:
                time.sleep(0.01)
                if not idle:
                    idle = True
                    self.logger.DEBUG("{} {} thread idle".format(self.__header, header))
            else:
                idle = False

            # 接入连接
            if not cnt % 5:
                con = self.__server.get_connection()
                if con is not None:
                    assert isinstance(con, Handler)
                    self.logger.INFO("{} {} incoming new connection ({}:{})".format(
                        self.__header, header, con.addr[0], con.addr[1]
                    ))
                    full_speed_ts = time.time()
                    con.send(b"I2FTP " + VERSION.encode())
                    self.connections.append(con)

            # 处理死亡的连接
            if not cnt % 100:
                self.connections = [con for con in self.connections if con.live]

            # 储存命令队列
            if not cnt % 2:
                for i, ele in enumerate(self.connections):
                    assert isinstance(ele, Handler)
                    package = ele.get()
                    if package is not None:
                        self.__cmd_queue.put((ele, package))
                        self.logger.DEBUG("{} {} \"{}\" request received, buffer size now {}".format(
                            self.__header, header, package[:4], self.__cmd_queue.qsize()
                    ))
            # 执行命令
            try:
                con, cmd = self.__cmd_queue.get(block=False)
            except queue.Empty:
                continue
            self.logger.DEBUG("{} {} [{}:{}] processing \"{}\" request".format(
                self.__header, header, con.addr[0], con.addr[1], cmd[0:4]
            ))
            try:
                assert isinstance(con, Handler)
                ret, ret_cmd = self.__process_requests(cmd)
                con.send(ret)
                if ret_cmd:
                    self.__cmd_queue.put((con, ret_cmd))
            except Exception as err:
                self.logger.ERROR("{} {} [{}:{}] failed to process requests of \"{}\", {}".format(
                    self.__header, header, con.addr[0], con.addr[1], cmd, err
                ))

                if DEBUG:
                    raise err

        self.threads_running["loop"] = False


def mannual():
    print("""I2FTP Server [by I2cy] v{}
    
Usage: i2ftps [-c CONFIG] [start/stop Default: start]

Options:
 -c --config CONFIG             - config JSON file path
                                  (default: /usr/share/i2ftp/
                                  server_conf.json)
 
 -h --help                      - display this message
 
Examples:
 >i2ftps
 >i2ftps -c server_conf.json
 >i2ftps -c server_conf.json start
 >i2ftps stop
 >i2ftps -c server_conf.json stop
    """.format(VERSION))


def main():
    if os.name == "nt":
        config_path = "C:/i2ftp/server_conf.json"
    else:
        config_path = "/usr/share/i2ftp/server_conf.json"

    opts = get_args()

    operation = "start"

    for opt in opts.keys():
        if opt in ("-c", "--config"):
            config_path = opts[opt]
        elif opt in ("-h", "--help"):
            mannual()
            return
        elif opt == 0:
            operation = opts[0]
            if operation not in ("start", "stop"):
                print("unhandled operation \"{}\"".format(operation))
                return

    config_path = pathlib.Path(config_path)
    if not config_path.exists():
        config = Config()
        config.dump(config_path)
    else:
        config = Config()
        config.load(config_path)

    if operation == "start":
        server = I2ftpServer(config)
        server.start()

        try:
            while True:
                cmd = input("")
                if cmd in ("quit", "stop", "kill"):
                    break
        except (KeyboardInterrupt, SystemExit):
            pass

        server.stop()

    else:
        pm.stop(config)


if __name__ == '__main__':
    main()



#!/usr/bin/python3
# -*- coding: utf-8 -*-
# Author: i2cy(i2cy@outlook.com)
# Filename: setup
# Created on: 2021/3/6

import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="i2ftps",  # Replace with your own username
    version="0.2.2",
    author="I2cy Cloud",
    author_email="i2cy@outlook.com",
    description="A FTP server based on I2TCP",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/i2cy/i2ftp-server",
    project_urls={
        "Bug Tracker": "https://github.com/i2cy/I2FTPServer/issues",
        "Source Code": "https://github.com/i2cy/I2FTPServer",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        'i2cylib >= 1.8.14'
    ],
    packages=setuptools.find_packages(),
    python_requires=">=3.6",
    entry_points={'console_scripts':
        [
            "i2ftps-setup = i2ftps.config:main",
            "i2ftps = i2ftps.server:main"
        ]
    }
)



#!/usr/bin/python3
# -*- coding: utf-8 -*-
# Author: i2cy(i2cy@outlook.com)
# Project: I2FTPServer
# Filename: config
# Created on: 2022/5/12


import json
import os
from i2cylib.utils.path import path_fixer


class Config(object):

    def __init__(self, filename=None):

        if os.name == "nt":
            default_root = r"C:/i2ftp"
        else:
            default_root = "/usr/share/i2ftp"

        self.ftp_root = "{}/root".format(default_root)  # FTP 根目录
        self.tls_enabled = False  # 启用I2TCP的传输层加密
        self.keychain = b"&90%]>__AdfI2FTP$F%_+@$^:aBasicKey%_+@-$^:>"  # 密钥
        self.read_only = False  # 服务器是否只读
        self.port = 26842  # 端口
        self.log_file = "{}/i2ftp.log".format(default_root)  # 日志文件
        self.log_level = "INFO"  # 日志等级（DEBUG、INFO、WARNING、ERROR、CRITICAL）

        if filename is not None:
            self.load(filename)

    def dump(self, filename):
        configs = {"root": self.ftp_root,
                   "port": self.port,
                   "log_file": self.log_file,
                   "log_level": self.log_level,
                   "keychain": self.keychain.hex(),
                   "tls": self.tls_enabled,
                   "read_only": self.read_only}

        with open(filename, "w") as f:
            json.dump(configs, f, indent=2)
            f.close()

    def load(self, filename):
        with open(filename, "r") as f:
            configs = json.load(f)
            f.close()
        self.ftp_root = configs["root"]
        self.port = configs["port"]
        self.tls_enabled = configs["tls"]
        self.read_only = configs["read_only"]
        self.keychain = bytes.fromhex(configs["keychain"])
        self.log_level = configs["log_level"]
        self.log_file = configs["log_file"]


def generate_systemd(config_filename, addon=None):
    if addon is None or addon == 1:
        addon = ""
    text = """[Unit]
Description=I2FTP Server Service
After=network.target

[Service]
Type=simple
DynamicUser=true
Restart=on-failure
RestartSec=20s
ExecStart=/usr/local/bin/i2ftps -c \"{}\" start
ExecStop=/usr/local/bin/i2ftps -c \"{}\" stop
LimitNOFILE=1048576

[Install]
WantedBy=multi-user.target
Alias=i2ftps{}.service
""".format(config_filename, config_filename, addon)

    return text


def main():
    print("Welcome to I2FTP Server Configurator [by I2cy]\n")
    print("this wizard will guide you to install I2FTP server"
          " or just generate a configuration file if you like."
          "\nNotice: you can always run this wizard by command "
          "\"i2ftps-setup\"")

    steps = 9
    if os.name == "nt":
        steps -= 1

    config = Config()

    # 设置传输目录根地址
    root = None
    while root is None:
        root = input("step 1/{}: Set the root path of I2FTP to share with\n"
                     "(input nothing for default: {}): ".format(steps, config.ftp_root))
        if not root:
            root = config.ftp_root

        try:
            if not os.path.exists(root):
                os.makedirs(root)
        except Exception as err:
            print("error: {}".format(err))
            root = None

    config.ftp_root = root

    # 设置端口
    port = None
    while port is None:
        port = input("step 2/{}: Set the port number of server to listen on\n"
                     "(input nothing for default: {}): ".format(steps, config.port))
        if not port:
            port = config.port

        try:
            port = int(port)
            if port < 1 or port > 65535:
                raise Exception("port must be greater than 0 and smaller than 65536")
        except Exception as err:
            print("error: {}".format(err))
            port = None

    config.port = port

    # 设置密钥
    key = None
    while key is None:
        key = input("step 3/{}: Set the token for server to authentication\n"
                    "(input nothing for default: \"{}\"): ".format(steps, config.keychain.decode()))
        if not key:
            key = config.keychain.decode()

        try:
            key = key.encode()
        except Exception as err:
            print("error: {}".format(err))
            key = None

    config.keychain = key

    # 设置日志文件
    log_file = None
    while log_file is None:
        log_file = input("step 4/{}: Set log's filename\n"
                         "(input nothing for default: {}): ".format(steps, config.log_file))
        if not log_file:
            log_file = config.log_file

        try:
            path_fixer(log_file)
        except Exception as err:
            print("error: {}".format(err))
            log_file = None

    config.log_file = log_file

    # 设置日志等级
    log_level = None
    while log_level is None:
        log_level = input("step 5/{}: Set logging level (DEBUG、INFO、WARNING、ERROR、CRITICAL)\n"
                          "(input nothing for default: {}): ".format(steps, config.log_level))
        if not log_level:
            log_level = config.log_level

        try:
            log_level = log_level.upper()
            if log_level not in ("DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"):
                raise Exception("logging level must be selected from DEBUG, INFO, WARNING, ERROR or CRITICAL")
        except Exception as err:
            print("error: {}".format(err))
            log_level = None

    config.log_level = log_level

    # 设置服务器是否只读
    readonly = None
    while readonly is None:
        readonly = input("step 6/{}: Read-Only server? (Y/N)\n"
                         "(input nothing for default: {}): ".format(steps, config.read_only))

        if readonly:
            if readonly in ("y", "Y", "Yes", "yes"):
                readonly = True
            else:
                readonly = False

        else:
            readonly = config.read_only

    config.read_only = readonly

    # 设置服务器是否加密
    tls = None
    while tls is None:
        tls = input("step 7/{}: Use TLS? (Y/N)\n"
                    "(input nothing for default: {}): ".format(steps, config.read_only))

        if tls:
            if tls in ("y", "Y", "Yes", "yes"):
                tls = True
            else:
                tls = False

        else:
            tls = config.read_only

    config.tls_enabled = tls

    # 导出配置文件至
    config_path = None
    if os.name == "nt":
        default_config = "C:/i2ftp/server_conf.json"
    else:
        default_config = "/usr/share/i2ftp/server_conf.json"

    while config_path is None:
        config_path = input("step 8/{}: Saving config file to...\n"
                            "(input nothing for default: {}): ".format(steps, default_config))

        if not config_path:
            config_path = default_config

        try:
            path_fixer(config_path)
            config.dump(config_path)
        except Exception as err:
            print("error: {}".format(err))
            config_path = None

    # 生成启动脚本
    if os.name != "nt":
        choice = None

        while choice is None:
            choice = input("step 9/{}: Add to systemd for autostart? (Y/N)\n"
                           "(input nothing for default: True): ".format(steps))

            if choice:
                if choice in ("y", "Y", "Yes", "yes"):
                    choice = True
                else:
                    choice = False

            else:
                choice = True

        if choice:
            try:
                addon = 1
                path = "/etc/systemd/system/i2ftps.service"
                while os.path.exists(path):
                    addon += 1
                    path = "/etc/systemd/system/i2ftps{}.service".format(addon)
                text = generate_systemd(config_path, addon)
                with open(path, "w") as f:
                    f.write(text)
                    f.close()
                if addon == 1:
                    addon = ""
                os.system("systemctl enable i2ftps{}".format(addon))
            except Exception as err:
                print("error: {}".format(err))


if __name__ == '__main__':
    main()